// VinciaTools.h is a part of the VINCIA plug-in to the PYTHIA 8 generator.
// Copyright (C) 2015 P. Skands.
// VINCIA is licenced under the GNU GPL version 2.
// Please respect the MCnet Guidelines for Event Generator Authors and Users, 
//               http://www.montecarlonet.org/GUIDELINES

// This file contains header information for VINCIA's Toolbox.
  
// Classes:
// VinClu : VINCIA's clustering maps
// Rambo  : Flat phase-space generator

#ifndef VINCIA_VinciaTools_H
#define VINCIA_VinciaTools_H

#include "Pythia8/Basics.h"             // Vec4
#include "VinciaCommon.h"       // Vec4toD4, m2(...)

using namespace Pythia8;

namespace Vincia {

  //**************************************************************************

  // VINCIA Clustering Map

  class VinClu {

  public:
    
    // Constructor.
    VinClu() {isInitPtr=false; isInit=false;}

    // Destructor.
    virtual ~VinClu() {}

    // Initialize pointers
    void initPtr(Rndm* rndmPtrIn, Settings* settingsPtrIn, 
		 VinciaCommon* vinComPtrIn) {
      rndmPtr     = rndmPtrIn; 
      settingsPtr = settingsPtrIn; 
      vinComPtr   = vinComPtrIn;
      isInitPtr   = true;
    }

    // Intialize VinClu (currently just to set verbose level)
    bool init() {
      verbose           = settingsPtr->mode("Vincia:verbose");
      isInit = true;
      return true;
    }

    // Function to cluster a set of partons
    // (basically vc3to2 overlaid with vectors of parton IDs from which
    //  the appropriate clustering is automatically determined and the 
    //  correct clustering map retrieved from the corresponding antenna)
    bool getClu(vector<int> idIn, vector<Vec4> pIn,int ia, int ir, int ib, 
		vector<int>& idClu, vector<Vec4>& pClu, 
		int& antIndx, bool& swapAB);
    
    // getCluPol uses integers to represent polarizations
    bool getCluPol(vector<int> idIn, vector<Vec4> pIn,int ia, int ir, int ib, 
                   vector<int>& idClu, vector<Vec4>& pClu, 
                   int& antIndx, bool& swapAB,
                   vector<int> polIn, vector< vector<int> >& polClu);
    // Wrapper to translate from double-precision input polarizations
    bool getCluPol(vector<int> idIn, vector<Vec4> pIn,int ia, int ir, int ib, 
		     vector<int>& idClu, vector<Vec4>& pClu, 
		     int& antIndx, bool& swapAB,
		     vector<double> polIn, vector< vector<int> >& polClu) {
      vector<int> polInInt(polIn.begin(), polIn.end());    
      copy(polIn.begin(), polIn.end(), back_inserter(polInInt));
      bool isOK = getCluPol(idIn, pIn, ia, ir, ib, idClu, pClu, antIndx, swapAB,
			    polInInt, polClu);
      return isOK;
    }
    
    // Main Vincia clustering map
    // MR 12.7.2010: changed to return bool
    bool vc3to2(int iMap,vector<Vec4>,int,int,int,vector<Vec4>&);
	
    // MR 7.7.2010: massive generalization of the clustering map, takes
    // the two parent masses as additional arguments
    // MR 12.7.2010: changed to return bool
    bool vc3to2massive(int iMap, vector<Vec4>, int, int, int, double,
		       double, vector<Vec4>&);
    
    // 2->3 branching kinematics
    // New structure: output arguments first, then input arguments
    // Output: p3
    // Input: p2, yij, yjk, phi, kineMapType
    bool map2to3(vector<Vec4>& pThree, vector<Vec4> pTwo, 
                 int kineMapTypeIn, 
                 double y12, double y23, double phi, 
                 double mu1=0.0, double mu2=0.0, double mu3=0.0);
    
    // 2->3 branching kinematics, massless
    // New structure: output arguments first, then input arguments
    // Output: p3
    // Input: p2, yij, yjk, phi, kineMapType
    bool map2to3light(vector<Vec4>& pThree, vector<Vec4> pTwo, 
                      int kineMapTypeIn, 
                      double y12, double y23, double phi);
    
    //******

    // Individual maps, accessed by the main algorithm
    // Imap = 1 & 2 so far accessed by VC3TO2 fortran routine
	
    // MR 11.7.2010: more general version accepting kineMapType
    // as the first argument, changed to return a boolean
    bool map3to2(int kineMapType, Vec4 pi, Vec4 pj, Vec4 pk,
		 Vec4& pOutI, Vec4& pOutK);
    
    // MR 7.7.2010: massive generalization of 3->2 clustering, takes
    // map type as the first argument
    bool map3to2massive(int kineMapType, Vec4 pi, Vec4 pj, Vec4 pk,
			double mI,double mK, Vec4& pOutI,Vec4& pOutK);

    // Check if 2-particle system is on-shell and rescale if not
    bool onShellCM(Vec4& p1, Vec4& p2, double m1, double m2, double tol=1e-6) {
      double s1     = pow2(m1);
      double s2     = pow2(m2);
      double s12    = Vec4(p1+p2).m2Calc();
      double s1Calc = p1.m2Calc();
      double s2Calc = p2.m2Calc();
      if (abs(s1Calc-s1)/s12 > tol || abs(s2Calc-s2)/s12 > tol ) {
	if (verbose >= 3) 
	  printOut("VinClu::onShellCM","forcing particles on mass shell");
	RotBstMatrix M;
	M.fromCMframe(p1,p2);
	// Define massive on-shell momenta
	double E1 = (s12 + s1 - s2)/(2*sqrt(s12));
	double E2 = (s12 - s1 + s2)/(2*sqrt(s12));
	double pz = pow2(E1)-s1;
	Vec4 p1new = Vec4(0.0,0.0,-pz,E1);
	Vec4 p2new = Vec4(0.0,0.0,pz,E2);
	p1new.rotbst(M);
	p2new.rotbst(M);
	double s1Test = p1new.m2Calc();
	double s2Test = p2new.m2Calc();
	if (verbose >= 3) {
	  cout<<" p1   : "<<p1<<" p1new: "<<p1new;
	  cout<<" p2   : "<<p1<<" p2new: "<<p1new;
	}
	// If this got them closer to mass shell, replace momenta
	if (abs(s1Test-s1)/s12 <= abs(s1Calc-s1)/s12
	    && abs(s2Test-s2)/s12 <= abs(s2Calc-s2)/s12) {
	  p1 = p1new;
	  p2 = p2new;
	}
	return false;
      }
      else return true;
    } 

    // isMassive()
    // Check if a set of momenta contains any numerically significant masses
    bool isMassive(const vector<Vec4>& momenta, 
                   int iChkFirst=0, int iChkLast=-1);

    // Wrapper to vector of particle pointers.
    // Read out 4-momenta, then call isMassive
    bool isMassive(const vector<Particle>& particles, 
                   int iChkFirst=0, int iChkLast=-1) {
      // Ensure sensible range for i
      iChkFirst = max(0,iChkFirst);
      iChkLast = min(iChkLast,int(particles.size())-1);
      // Default value (-1) means up to last element
      if (iChkLast < 0) iChkLast = particles.size()-1;
      // Fill four-momentum vector
      vector<Vec4> momenta;
      for (int i=iChkFirst; i<=iChkLast; ++i) 
	momenta.push_back(particles[i].p());
      return isMassive(momenta,0,momenta.size()-1);
    }

    // isRelativistic():
    // Returns true if a set of partons can be treated as relativistic
    // The set of partons to be tested is defined by the range
    // [iChkFirst,iChkLast] inside a possibly larger vector of momenta
    // relativisticSpeed: set speed beyond which particles can be made
    // massless. Default (0.) is to always do the mapping, but when
    // called from VINCIA, a non-zero value will be used. The method
    // then returns false if the remapping cannot be done. 
    bool isRelativistic(const vector<Vec4>& momenta, 
                        const double relativisticSpeed,
                        int iChkFirst=0, int iChkLast=-1);

    // Wrapper to vector of particle pointers.
    // Read out 4-momenta, then call isRelativistic
    bool isRelativistic(const vector<Particle>& particles, 
                        const  double relativisticSpeed,
                        int iChkFirst=0, int iChkLast=-1) {
      // Ensure sensible range for i
      iChkFirst = max(0,iChkFirst);
      iChkLast = min(iChkLast,int(particles.size())-1);
      // Default value (-1) means up to last element
      if (iChkLast == -1) iChkLast = particles.size()-1;
      // Fill four-momentum vector
      vector<Vec4> momenta;
      for (int i=iChkFirst; i<=iChkLast; ++i) 
	momenta.push_back(particles[i].p());
      return isRelativistic(momenta, relativisticSpeed, 0, momenta.size()-1);
    }

    // mapToMassless():
    // Map a set of massive momenta to a set of equivalent massless ones
    // Boolean returns true if a mapping was made, else false
    // iFirst: ignore particles before iFirst (e.g., resonance mother)
    bool mapToMassless(vector<Vec4>& momenta, int iFirst=0);

    // Obsolete: old mapToMassless 
    bool mapToMasslessOld(vector<Vec4>& momenta, int iFirst=0);

    // Wrapper to vector of particle pointers.
    // Read out 4-momenta, call mapToMassless, then save new momenta.
    bool mapToMassless(vector<Particle>& particles, int iFirst=0) {
      vector<Vec4> momenta;
      int nParticles=particles.size();
      // Ensure sensible range for i
      iFirst = max(0,iFirst);
      for (int i=iFirst; i<nParticles; ++i) 
	momenta.push_back(particles[i].p());
      if (mapToMassless(momenta,0)) {
        // Save massless momenta (remembering that PYTHIA particles have
        // a separate component for storing the mass value)
	for (int i=iFirst; i<nParticles; ++i) {
	  particles[i].p(momenta[i-iFirst]);
          particles[i].m(0.0);
        }
        return true;
      }
      return false;
    }

    // Map a massless antenna to equivalent massive one
    // Boolean returns true if a modification was made
    bool map2massive(Vec4& p1, Vec4& p2, double m1, double m2) {
      return (!onShellCM(p1,p2,m1,m2,1e-9));      
    }
    
    // Get/Set parameters
    int getVerbose() {return verbose;}
    void setVerbose(int verboseIn) {verbose=verboseIn;}

  private :
  
    // Is initialized?
    bool isInitPtr, isInit;

    // Pointers to VINCIA and PYTHIA 8 objects
    Rndm*  rndmPtr;
    Settings* settingsPtr;
    VinciaCommon* vinComPtr;

    // Verbose level    
    int verbose;

  };

  //**************************************************************************

  // Rambo flat phase-space generator

  class Rambo {
   
  public:
    
    // Deafult Constructor.
    Rambo() {isInitPtr=false;}
    
    // Initializing Constructor.
    Rambo(Rndm* rndmPtrIn) { initPtr(rndmPtrIn); }

    // Destructor.
    virtual ~Rambo() {}
    
    // Initialize pointers
    void initPtr(Rndm* rndmPtrIn) {rndmPtr = rndmPtrIn; isInitPtr=true;}
    
    // Rambo phase space generator
    // p[0][i] is the energy of the i'th parton
    double genPoint(double S,int N,double p[4][8]);
    // MR 6.2.2011: massive generalisation, weights NOT 1 anymore
    // literal implementation of original RAMBO paper by Ellis, 
    // Kleiss and Stirling
    double genPoint(double S,int N,double m[8],double p[4][8]);
    // MR 15.3.2011: implement a function which returns a sequence 
    // of massless vectors in which the n1 and n2 become 
    // collinear
    vector< vector<Vec4> > collinearSequence(
      double S, int nParticles, int nSteps, int n1, int n2, 
      double rescalingFactor, double rescalingFactorTolerance = 1e-3 );
    // MR 28.2011: re-implement the RAMBO algorithm according to 
    // the original paper again as a check
    double genPointPrime(
      double ecm, int n, vector<double> masses, vector<Vec4>& retVec);
    
  private:
    
    // Is initialized?
    bool isInitPtr;

    // Pointer to the random number generator.
    Rndm*  rndmPtr;
    
  };

}

#endif
